// Copyright (C) 2021 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include <QtQuick3DRuntimeRender/private/qssgrenderreflectionprobe_p.h>

QT_BEGIN_NAMESPACE

QSSGRenderReflectionProbe::QSSGRenderReflectionProbe()
    : QSSGRenderNode(QSSGRenderGraphObject::Type::ReflectionProbe)
{
}

QT_END_NAMESPACE
