// Copyright (C) 2022 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause

/*
 * This file is automatically generated by Qt Design Studio.
 * Do not change.
*/

#include <QtQml/qqmlextensionplugin.h>

Q_IMPORT_QML_PLUGIN(contentPlugin)
Q_IMPORT_QML_PLUGIN(RobotArmPlugin)
