// Copyright (C) 2019 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#ifndef QTQUICK3DGLOBAL_H
#define QTQUICK3DGLOBAL_H

#include <QtCore/qglobal.h>
#include <QtQuick3D/qtquick3dexports.h>

#endif // QTQUICK3DGLOBAL_H
