// Copyright (C) 2020 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only


#include <QtQuick3DRuntimeRender/private/qssgrenderskeleton_p.h>

QT_BEGIN_NAMESPACE

QSSGRenderSkeleton::QSSGRenderSkeleton()
    : QSSGRenderNode(QSSGRenderGraphObject::Type::Skeleton)
{
}

QSSGRenderSkeleton::~QSSGRenderSkeleton() = default;

QT_END_NAMESPACE
