// Copyright (C) 2022 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include "qssgrenderskin_p.h"

QT_BEGIN_NAMESPACE

QSSGRenderSkin::QSSGRenderSkin()
    : QSSGRenderGraphObject(QSSGRenderGraphObject::Type::Skin)
{
}

QSSGRenderSkin::~QSSGRenderSkin()
{
}

QT_END_NAMESPACE
