/* This file was generated by syncqt. */
#ifndef QT_QTQUICK3DASSETUTILS_VERSION_H
#define QT_QTQUICK3DASSETUTILS_VERSION_H

#define QTQUICK3DASSETUTILS_VERSION_STR "6.4.2"

#define QTQUICK3DASSETUTILS_VERSION 0x060402

#endif // QT_QTQUICK3DASSETUTILS_VERSION_H
